# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.editor
"""

BENCHMARK = True

import logging
if BENCHMARK: import time

from gi.repository import GObject

from ..editor import Editor
from ..file import File
from ..issues import Issue, IIssueHandler

from .parser import LaTeXParser
from .expander import LaTeXReferenceExpander
from .outline import LaTeXOutlineGenerator
from .validator import LaTeXValidator
from .completion import LaTeXCompletionHandler

from .dialogs import ChooseMasterDialog

from . import LaTeXSource
from ..preferences import Preferences, DocumentPreferences

LOG = logging.getLogger(__name__)

class LaTeXEditor(Editor, IIssueHandler):

    extensions = Preferences().get("latex-extensions").split(",")

    dnd_extensions = [".png", ".pdf", ".bib", ".tex"]

    @property
    def completion_handlers(self):
        self.__latex_completion_handler = LaTeXCompletionHandler()
        return [ self.__latex_completion_handler ]

    def init(self, file, context):
        """
        @param file: base.File
        @param context: base.WindowContext
        """
        LOG.debug("init(%s)" % file)

        self._file = file
        self._context = context

        self._preferences = DocumentPreferences(self._file)
        self._preferences.connect("preferences-changed", self._on_preferences_changed)

        style_scheme = self._text_buffer.get_style_scheme()
        w_style = style_scheme.get_style('def:warning')
        w_color = w_style.get_properties('background')[0]
        self.register_marker_type("latex-warning", w_color)

        e_style = style_scheme.get_style('def:error')
        e_color = e_style.get_properties('background')[0]
        self.register_marker_type("latex-error", e_color)

        self._issue_view = context.find_view(self, "IssueView")
        self._outline_view = context.find_view(self, "LaTeXOutlineView")

        self._parser = LaTeXParser()
        self._outline_generator = LaTeXOutlineGenerator()
        self._validator = LaTeXValidator()
        self._document = None

        # if the document is no master we display an info message on the packages to
        # include - _ensured_packages holds the already mentioned packages to not
        # annoy the user
        self._ensured_packages = []

        #
        # initially parse
        #
        self._change_reference = self.initial_timestamp

        GObject.idle_add(self.__parse)
        self.__update_neighbors()

    def _on_preferences_changed(self, prefs, key, new_value):
        if key in ["outline-show-labels", "outline-show-tables", "outline-show-graphics"]:
            # regenerate outline model
            if self._document_is_master:
                self._outline = self._outline_generator.generate(self._document, self)
                self._outline_view.set_outline(self._outline)
            else:
                # FIXME: self._document contains the full model of child and master
                # so we may not use it for regenerating the outline here
                self.__parse()

    def drag_drop_received(self, files):
        # see base.Editor.drag_drop_received

        # TODO: we need to insert the source at the drop location - so pass it here

        LOG.debug("drag_drop: %s" % files)

#        if len(files) == 1:
#            file = files[0]
#            self._log.debug("Got one file: %s, extension: %s" % (file.path, file.extension))
#            if file.extension == ".png":
#                self._log.debug("PNG image - including...")
#                source = "\\includegraphics{%s}" % file.path
#                self.insert(source)

    def insert(self, source):
        # see base.Editor.insert

        if type(source) is LaTeXSource:
            if source.packages and len(source.packages) > 0:
                self.ensure_packages(source.packages)

            Editor.insert(self, source.source)
        else:
            Editor.insert(self, source)

    POSITION_PACKAGES, POSITION_BIBLIOGRAPHY = 1, 2

    def insert_at_position(self, source, position):
        """
        Insert source at a certain position in the LaTeX document:

         * POSITION_PACKAGES: after the last usepackage statement
         * POSITION_BIBLIOGRAPHY: before \end{document}

        @param source: a LaTeXSource object
        @param position: POSITION_PACKAGES | POSITION_BIBLIOGRAPHY
        """

        if position == self.POSITION_BIBLIOGRAPHY:
            offset = self._document.end_of_document
            Editor.insert_at_offset(self, offset, source, True)
        elif position == self.POSITION_PACKAGES:
            offset = self._document.end_of_packages
            Editor.insert_at_offset(self, offset, source, False)
        else:
            raise NotImplementedError

    def ensure_packages(self, packages):
        """
        Ensure that certain packages are included

        @param packages: a list of package names
        """
        self.__parse()    # ensure up-to-date document model

        if not self._document_is_master:
            LOG.debug("ensure_packages: document is not a master")

            # find the packages that haven't already been mentioned
            info_packages = [p for p in packages if not p in self._ensured_packages]

            if len(info_packages) > 0:
                # generate markup
                li_tags = "\n".join([" • <tt>%s</tt>" % p for p in info_packages])

                from ..util import open_info
                open_info("LaTeX Package Required",
                        "Please make sure that the following packages are included in the master document per <tt>\\usepackage</tt>: \n\n%s" % li_tags)

                # extend the already mentioned packages
                self._ensured_packages.extend(info_packages)

            return

        # find missing packages
        present_package_names = [p.value for p in self._outline.packages]
        package_names = [p for p in packages if not p in present_package_names]

        # insert the necessary \usepackage commands
        if len(package_names) > 0:
            source = "\n" + "\n".join(["\\usepackage{%s}" % n for n in package_names])
            self.insert_at_position(source, self.POSITION_PACKAGES)

    def on_save(self):
        """
        The file has been saved

        Update models
        """
        LOG.debug("document saved") 
       
#        from multiprocessing import Process
#
#        p_parse = Process(target=self.__parse)
#        p_parse.start()

        self.__parse()

    def __update_neighbors(self):
        """
        Find all files in the working directory that are relevant for LaTeX, e.g.
        other *.tex files or images.
        """

        # TODO: this is only needed to feed the LaTeXCompletionHandler. So maybe it should
        # know the edited file and the Editor should call an update() method of the handler
        # when the file is saved.

        #FIXME: Look in graphicspath here, and in subdirs.

        tex_files = self._file.find_neighbors(".tex")
        bib_files = self._file.find_neighbors(".bib")

        graphic_files = []
        for extension in [".ps", ".pdf", ".png", ".jpg", ".eps"]:
            graphic_files.extend(self._file.find_neighbors(extension))

        self.__latex_completion_handler.set_neighbors(tex_files, bib_files, graphic_files)

    #@verbose
    def __parse(self):
        """
        Ensure that the document model is up-to-date
        """
        if self.content_changed(self._change_reference):
            # content has changed so document model may be dirty
            self._change_reference = self.current_timestamp

            LOG.debug("Parsing document...")

            # reset highlight
            self.remove_markers("latex-error")
            self.remove_markers("latex-warning")

            # reset issues
            self._issue_view.clear()

            if BENCHMARK: t = time.clock()

            # parse document
            if self._document != None:
                self._document.destroy()
                del self._document
            self._document = self._parser.parse(self.content, self._file, self)

            # update document preferences
            self._preferences.parse_content(self.content)

            if BENCHMARK: LOG.info("LaTeXParser.parse: %f" % (time.clock() - t))

            LOG.debug("Parsed %s bytes of content" % len(self.content))

            # FIXME: the LaTeXChooseMasterAction enabled state has to be updated on tab change, too!

            if self._document.is_master:

                self._context.set_action_enabled("LaTeXChooseMasterAction", False)
                self._document_is_master = True

                # expand child documents
                expander = LaTeXReferenceExpander()
                expander.expand(self._document, self._file, self, self.charset)

                # generate outline from the expanded model
                self._outline = self._outline_generator.generate(self._document, self)

                # pass to view
                self._outline_view.set_outline(self._outline)

                # validate
                self._validator.validate(self._document, self._outline, self, self._preferences)
            else:
                LOG.debug("Document is not a master")

                self._context.set_action_enabled("LaTeXChooseMasterAction", True)
                self._document_is_master = False

                # the outline used by the outline view has to be created only from the child model
                # otherwise we see the outline of the master and get wrong offsets
                self._outline = self._outline_generator.generate(self._document, self)
                self._outline_view.set_outline(self._outline)

                # find master
                master_file = self.__master_file

                if master_file is None:
                    return

                # parse master
                master_content = open(master_file.path).read()
                self._document = self._parser.parse(master_content, master_file, self)

                # expand its child documents
                expander = LaTeXReferenceExpander()
                expander.expand(self._document, master_file, self, self.charset)

                # create another outline of the expanded master model to make elements
                # from the master available (labels, colors, BibTeX files etc.)
                self._outline = self._outline_generator.generate(self._document, self)

                # validate
                prefs = DocumentPreferences(master_file)
                prefs.parse_content(master_content)
                self._validator.validate(self._document, self._outline, self, prefs)

            # pass outline to completion
            self.__latex_completion_handler.set_outline(self._outline)

            # pass neighbor files to completion
            self.__update_neighbors()

            LOG.debug("Parsing finished")

    def choose_master_file(self):
        master_filename = ChooseMasterDialog().run(self._file.dirname)
        if master_filename:
            # relativize the master filename
            master_filename = File(master_filename).relativize(self._file.dirname, True)
            self._preferences.set("master-filename", master_filename)
        return master_filename

    @property
    def __master_file(self):
        """
        Find the LaTeX master of this child

        @return: base.File
        """
        path = self._preferences.get("master-filename")
        if path != None:
            if File.is_absolute(path):
                LOG.debug("master path is absolute")
                return File(path)
            else:
                LOG.debug("master path is relative")
                return File.create_from_relative_path(path, self._file.dirname)
        else:
            # master filename not found, ask user
            master_filename = self.choose_master_file()
            if master_filename:
                return File.create_from_relative_path(master_filename, self._file.dirname)
            else:
                return None

    def issue(self, issue):
        # see IIssueHandler.issue

        local = (issue.file == self._file)

        self._issue_view.append_issue(issue, local)

        if issue.file == self._file:
            if issue.severity == Issue.SEVERITY_ERROR:
                self.create_marker("latex-error", issue.start, issue.end)
            elif issue.severity == Issue.SEVERITY_WARNING:
                self.create_marker("latex-warning", issue.start, issue.end)

    def on_cursor_moved(self, offset):
        """
        The cursor has moved
        """
        if self._preferences.get("outline-connect-to-editor"):
            self._outline_view.select_path_by_offset(offset)

    @property
    def file(self):
        # overrides Editor.file

        # we may not call self._document.is_master because _document is always
        # replaced by the master model
        if self._document_is_master:
            return self._file
        else:
            return self.__master_file

    @property
    def edited_file(self):
        """
        Always returns the really edited file instead of the master

        This is called by the outline view to identify foreign nodes
        """
        return self._file

    def destroy(self):
        # unreference the window context
        del self._context

        # destroy the cached document
        self._document.destroy()
        del self._document

        Editor.destroy(self)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.validator
"""
from logging import getLogger

from ..issues import Issue
from .model import BibTeXModel


class BibTeXValidator:
    """
    This checks for
     - duplicate entry names
     - duplicate fields
     - missing required fields *
     - unused fields *

    *) relies on the BibTeX model definition
    """

    _log = getLogger("BibTeXValidator")

    def __init__(self):
        self._model = BibTeXModel()

    def validate(self, document, file, issue_handler):
        """
        @param document: a bibtex.parser.Document object
        @param issue_handler: an object implementing IIssueHandler
        """
        entry_keys = []
        for entry in document.entries:
            # check for duplicate keys
            if entry.key in entry_keys:
                issue_handler.issue(Issue("Duplicate key <b>%s</b>" % entry.key,
                                          entry.start, entry.end, file,
                                          Issue.SEVERITY_ERROR))
            else:
                entry_keys.append(entry.key)

            field_names = []
            for field in entry.fields:
                # check for duplicate fields
                if field.name in field_names:
                    issue_handler.issue(Issue("Duplicate field <b>%s</b>" % field.name,
                                        entry.start, entry.end, file, Issue.SEVERITY_ERROR))
                else:
                    field_names.append(field.name)

            try:
                # check for missing required fields
                required_field_names = set([f.name for f in self._model.find_type(entry.type).required_fields])
                missing_field_names = required_field_names.difference(set(field_names))

                if len(missing_field_names) > 0:
                    issue_handler.issue(Issue("Possibly missing field(s): <b>%s</b>" % ",".join(missing_field_names),
                                              entry.start, entry.end, file, Issue.SEVERITY_WARNING))

                # check for unused fields
                optional_field_names = set([f.name for f in self._model.find_type(entry.type).optional_fields])
                unused_field_names = set(field_names).difference(optional_field_names.union(required_field_names))

                if len(unused_field_names) > 0:
                    issue_handler.issue(Issue("Possibly unused field(s): <b>%s</b>" % ",".join(unused_field_names),
                                              entry.start, entry.end, file, Issue.SEVERITY_WARNING))
            except KeyError:
                #self._log.debug("Type not found: %s" % entry.type)
                pass

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
preferences.dialog
"""

from logging import getLogger
from gi.repository import Gtk

from ..resources import Resources
from ..util import GladeInterface
from ..tools import Tool, Job

from . import Preferences
from .tools import ToolPreferences

def _insert_column_with_attributes(view, pos, title, rend, **kwargs):
    tv = Gtk.TreeViewColumn(title)
    tv.pack_start(rend, True)
    for k in kwargs:
        tv.add_attribute(rend, k, kwargs[k])
    view.insert_column(tv, pos)

class PreferencesSpinButtonProxy(object):
    def __init__(self, widget, key):
        """
        @param widget: a SpinButton widget
        @param key:
        @param default_value:
        """
        self._widget = widget
        self._key = key
        self._preferences = Preferences()

        self._widget.set_value(int(self._preferences.get(key)))

        self._widget.connect("value-changed", self._on_value_changed)

    def _on_value_changed(self, spin_button):
        self._preferences.set(self._key, spin_button.get_value_as_int())

class ConfigureToolDialog(GladeInterface):
    """
    Wraps the dialog for setting up a Tool
    """

    _dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("configure_tool.ui")

    def run(self, tool):
        """
        Runs the dialog and returns the updated Tool or None on abort
        """
        dialog = self._get_dialog()

        self._tool = tool
        

        # load Tool
        self._entry_label.set_text(tool.label)
        self._entry_description.set_text(tool.description)

        self._store_job.clear()
        for job in tool.jobs:
            self._store_job.append([job.command_template, job.must_succeed, job.post_processor.name])

        self._store_extension.clear()
        for ext in tool.extensions:
            self._store_extension.append([ext])

        if tool.accelerator is None:
            self._radio_user_accel.set_active(False)
            self._entry_accel.set_text("")
        else:
            self._radio_user_accel.set_active(True)
            self._entry_accel.set_text(tool.accelerator)

        if dialog.run() == 1:
            #
            # okay clicked - update the Tool object
            #
            tool.label = self._entry_label.get_text()
            tool.description = self._entry_description.get_text()

            tool.jobs = []
            for row in self._store_job:
                pp_class = self._tool_preferences.POST_PROCESSORS[row[2]]
                tool.jobs.append(Job(row[0], row[1], pp_class))

            tool.extensions = []
            for row in self._store_extension:
                tool.extensions.append(row[0])

            # TODO: validate accelerator!
            if self._radio_user_accel.get_active():
                tool.accelerator = self._entry_accel.get_text()
            else:
                tool.accelerator = None

            return tool
        else:
            return None

    def _get_dialog(self):
        if not self._dialog:
            #
            # build the dialog
            #
            self._tool_preferences = ToolPreferences()

            self._dialog = self.find_widget("dialogConfigureTool")
            self._button_okay = self.find_widget("buttonOkay")
            self._labelProfileValidate = self.find_widget("labelHint")

            #
            # label
            #
            self._entry_label = self.find_widget("entryLabel")

            #
            # jobs
            #
            self._entry_new_job = self.find_widget("entryNewJob")
            self._button_add_job = self.find_widget("buttonAddJob")
            self._button_remove_job = self.find_widget("buttonRemoveJob")
            self._button_job_up = self.find_widget("buttonMoveUpJob")
            self._view_job = self.find_widget("treeviewJob")

            self._store_job = Gtk.ListStore(str, bool, str)   # command, mustSucceed, postProcessor

            self._view_job.set_model(self._store_job)

            mustSucceedRenderer = Gtk.CellRendererToggle()
            mustSucceedRenderer.connect("toggled", self._on_must_succeed_toggled)

            commandRenderer = Gtk.CellRendererText()
            commandRenderer.connect("edited", self._on_job_command_edited)

            self._store_pp = Gtk.ListStore(str)
            for p in self._tool_preferences.POST_PROCESSORS.keys():
                self._store_pp.append([p])

            ppRenderer = Gtk.CellRendererCombo()
            ppRenderer.set_property("editable", True)
            ppRenderer.set_property("model", self._store_pp)
            ppRenderer.set_property("text_column", 0)
            ppRenderer.set_property("has_entry", False)

            ppRenderer.connect("edited", self._on_job_pp_edited)

            _insert_column_with_attributes(self._view_job, -1, "Command", commandRenderer, text=0, editable=True)
            _insert_column_with_attributes(self._view_job, -1, "Must Succeed", mustSucceedRenderer, active=1)
            _insert_column_with_attributes(self._view_job, -1, "Post-Processor", ppRenderer, text=2)

            #
            # description
            #
            self._entry_description = self.find_widget("entryDescription")

            #
            # extensions
            #
            self._entry_new_extension = self.find_widget("entryNewExtension")

            self._store_extension = Gtk.ListStore(str)

            self._view_extension = self.find_widget("treeviewExtension")
            self._view_extension.set_model(self._store_extension)
            _insert_column_with_attributes(self._view_extension, -1, "", Gtk.CellRendererText(), text=0)
            self._view_extension.set_headers_visible(False)

            self._button_add_extension = self.find_widget("buttonAddExtension")
            self._button_remove_extension = self.find_widget("buttonRemoveExtension")

            self._radio_user_accel = self.find_widget("radioAccelUser")
            self._entry_accel = self.find_widget("entryAccel")

            self.connect_signals({ "on_entryNewJob_changed" : self._on_new_job_changed,
                                   "on_entryNewExtension_changed" : self._on_new_extension_changed,
                                   "on_buttonAddJob_clicked" : self._on_add_job_clicked,
                                   "on_buttonRemoveJob_clicked" : self._on_remove_job_clicked,
                                   "on_treeviewJob_cursor_changed" : self._on_job_cursor_changed,
                                   "on_treeviewExtension_cursor_changed" : self._on_extension_cursor_changed,
                                   "on_buttonAbort_clicked" : self._on_abort_clicked,
                                   "on_buttonOkay_clicked" : self._on_okay_clicked,
                                   "on_buttonRemoveExtension_clicked" : self._on_remove_extension_clicked,
                                   "on_buttonAddExtension_clicked" : self._on_add_extension_clicked,
                                   "on_buttonMoveUpJob_clicked" : self._on_move_up_job_clicked,
                                   "on_radioAccelUser_toggled" : self._on_accel_user_toggled })

        return self._dialog

    def _on_accel_user_toggled(self, togglebutton):
        enabled = togglebutton.get_active()
        self._entry_accel.set_sensitive(enabled)

    def _on_move_up_job_clicked(self, button):
        store, iter = self._view_job.get_selection().get_selected()
        prev = store.iter_previous(iter)
        if prev:
            store.swap(iter, prev)

    def _on_add_extension_clicked(self, button):
        extension = self._entry_new_extension.get_text()
        self._store_extension.append([extension])

    def _on_remove_extension_clicked(self, button):
        store, it = self._view_extension.get_selection().get_selected()
        store.remove(it)

    def _on_job_command_edited(self, renderer, path, text):
        """
        The command template has been edited
        """
        self._store_job.set_value(self._store_job.get_iter_from_string(path), 0, text)

    def _on_job_pp_edited(self, renderer, path, text):
        """
        Another post processor has been selected
        """
        self._store_job.set_value(self._store_job.get_iter_from_string(path), 2, text)

    def _on_must_succeed_toggled(self, renderer, path):
        """
        The 'must succeed' flag has been toggled
        """
        value = self._store_job.get(self._store_job.get_iter_from_string(path), 1)[0]
        self._store_job.set_value(self._store_job.get_iter_from_string(path), 1, not value)

    def _on_add_job_clicked(self, button):
        """
        Add a new job
        """
        command = self._entry_new_job.get_text()
        self._store_job.append([command, True, "GenericPostProcessor"])

    def _on_remove_job_clicked(self, button):
        store, it = self._view_job.get_selection().get_selected()
        store.remove(it)

    def _on_new_job_changed(self, widget):
        """
        The entry for a new command template has changed
        """
        self._button_add_job.set_sensitive(len(self._entry_new_job.get_text()) > 0)

    def _on_new_extension_changed(self, widget):
        self._button_add_extension.set_sensitive(len(self._entry_new_extension.get_text()) > 0)

    def _on_job_cursor_changed(self, tree_view):
        selection = tree_view.get_selection()
        if selection:
            store, iter = tree_view.get_selection().get_selected()
        if not selection or not iter:
            return
        self._button_remove_job.set_sensitive(True)

        first_row_selected = (store.get_string_from_iter(iter) == "0")
        self._button_job_up.set_sensitive(not first_row_selected)

    def _on_extension_cursor_changed(self, tree_view):
        selection = tree_view.get_selection()
        if selection:
            store, it = tree_view.get_selection().get_selected()
        if not selection or not it:
            return
        self._button_remove_extension.set_sensitive(True)

    def _on_abort_clicked(self, button):
        self._dialog.hide()

    def _on_okay_clicked(self, button):
        self._dialog.hide()

    def _validate_tool(self):
        """
        Validate the dialog contents
        """
        errors = []

        if len(self._store_job) == 0:
            errors.append("You have not specified any jobs.")

        if len(errors):
            self._buttonApply.set_sensitive(False)
        else:
            self._buttonApply.set_sensitive(True)

        if len(errors) == 1:
            self._labelProfileValidate.set_markup(errors[0])
        elif len(errors) > 1:
            self._labelProfileValidate.set_markup("\n".join([" * %s" % e for e in errors]))
        else:
            self._labelProfileValidate.set_markup("Remember to run all commands in batch mode (e.g. append <tt>-interaction batchmode</tt> to <tt>latex</tt>)")



class PreferencesDialog(GladeInterface):
    """
    This controls the configure dialog
    """

    _log = getLogger("PreferencesWizard")

    _dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("configure.ui")

    @property
    def dialog(self):
        if not self._dialog:
            self._dialog = self.find_widget("notebook1")

            self._preferences = Preferences()
            self._tool_preferences = ToolPreferences()

            # grab widgets

            self._button_delete_tool = self.find_widget("buttonDeleteTool")
            self._button_move_up_tool = self.find_widget("buttonMoveUpTool")
            self._button_configure_tool = self.find_widget("buttonConfigureTool")

            self._store_tool = Gtk.ListStore(str, str, object)     # label markup, extensions, Tool instance

            self._view_tool = self.find_widget("treeviewProfiles")
            self._view_tool.set_model(self._store_tool)
            _insert_column_with_attributes(self._view_tool, -1, "Label", Gtk.CellRendererText(), markup=0)
            _insert_column_with_attributes(self._view_tool, -1, "File Extensions", Gtk.CellRendererText(), text=1)

            # init tool and listen to tool changes
            self._load_tools()
            self._tool_preferences.connect("tools-changed", self._on_tools_changed)


            # misc
            check_hide_box = self.find_widget("checkHideBox")
            check_hide_box.set_active(self._preferences.get("hide-box-warnings"))


            #
            # proxies for SpinButtons
            #
            self._proxies = [PreferencesSpinButtonProxy(self.find_widget("spinMaxBibSize"), "maximum-bibtex-size")]

            #
            # signals
            #
            self.connect_signals({ "on_buttonClose_clicked" : self._on_close_clicked,
                                   "on_treeviewProfiles_cursor_changed" : self._on_tool_cursor_changed,
                                   "on_buttonAddBib_clicked" : self._on_add_bib_clicked,
                                   "on_buttonRemoveBib_clicked" : self._on_delete_bib_clicked,
                                   "on_buttonNewProfile_clicked" : self._on_new_tool_clicked,
                                   "on_buttonMoveUpTool_clicked" : self._on_tool_up_clicked,
                                   "on_buttonMoveDownTool_clicked" : self._on_tool_down_clicked,
                                   "on_buttonConfigureTool_clicked" : self._on_configure_tool_clicked,
                                   "on_buttonDeleteTool_clicked" : self._on_delete_tool_clicked,
                                   "on_checkHideBox_toggled" : self._on_hide_box_toggled})

        return self._dialog

    def _on_hide_box_toggled(self, togglebutton):
        value = togglebutton.get_active()
        self._preferences.set("hide-box-warnings", value)

    def _on_tools_changed(self, tools):
        self._load_tools()

    def _load_tools(self):
        # save cursor
        selection = self._view_tool.get_selection()
        if selection:
            store, iter = self._view_tool.get_selection().get_selected()
        if not selection or iter is None:
            restore_cursor = False
        else:
            path = store.get_path(iter)
            restore_cursor = True

        # reload tools
        self._store_tool.clear()
        for tool in self._tool_preferences.tools:
            self._store_tool.append(["<b>%s</b>" % tool.label, ", ".join(tool.extensions), tool])

        # restore cursor
        if restore_cursor:
            self._view_tool.set_cursor(path, None, False)

    def _on_configure_tool_clicked(self, button):
        store, it = self._view_tool.get_selection().get_selected()
        tool = store.get_value(it, 2)

        dialog = ConfigureToolDialog()

        if not dialog.run(tool) is None:
            self._tool_preferences.save_or_update_tool(tool)

    def _on_delete_tool_clicked(self, button):
        store, it = self._view_tool.get_selection().get_selected()
        tool = store.get_value(it, 2)

        self._tool_preferences.delete_tool(tool)

    def _on_tool_up_clicked(self, button):
        """
        Move up the selected tool
        """
        store, iter = self._view_tool.get_selection().get_selected()
        tool_1 = store.get_value(iter, 2)

        index_previous = int(store.get_string_from_iter(iter)) - 1
        assert index_previous >= 0

        iter_previous = store.get_iter_from_string(str(index_previous))
        tool_2 = store.get_value(iter_previous, 2)

        self._tool_preferences.swap_tools(tool_1, tool_2)

        # adjust selection
        self._view_tool.get_selection().select_path(str(index_previous))

    def _on_tool_down_clicked(self, button):
        """
        Move down the selected tool
        """
        store, iter = self._view_tool.get_selection().get_selected()
        tool_1 = store.get_value(iter, 2)

        index_next = int(store.get_string_from_iter(iter)) + 1
        assert index_next < len(store)

        iter_next = store.get_iter_from_string(str(index_next))
        tool_2 = store.get_value(iter_next, 2)

        # swap tools in preferences
        self._tool_preferences.swap_tools(tool_1, tool_2)

        # adjust selection
        self._view_tool.get_selection().select_path(str(index_next))

    def _on_new_tool_clicked(self, button):
        dialog = ConfigureToolDialog()

        tool = Tool("New Tool", [], "", None, [".tex"])

        if not dialog.run(tool) is None:
            self._tool_preferences.save_or_update_tool(tool)

    def _on_delete_bib_clicked(self, button):
        pass

    def _on_add_bib_clicked(self, button):
        pass

    def _on_close_clicked(self, button):
        self._dialog.hide()

    def _on_tool_cursor_changed(self, treeView):
        """
        The cursor in the tools view has changed
        """
        selection = treeView.get_selection()
        if selection:
            store, it = treeView.get_selection().get_selected()
        if not selection or not it:
            return

        self._profile = store.get_value(it, 1)

        self._button_delete_tool.set_sensitive(True)
        self._button_move_up_tool.set_sensitive(True)
        self._button_configure_tool.set_sensitive(True)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
base.decorators

These classes are 'attached' to the according Gedit objects. They form the
extension point.
"""

import logging

from gi.repository import GObject

from .config import EDITORS
from .file import File

# TODO: maybe create ActionDelegate for GeditWindowDecorator

LOG = logging.getLogger(__name__)

class GeditTabDecorator(object):
    """
    This monitors the opened file and manages the Editor objects
    according to the current file extension
    """

    def __init__(self, window_decorator, tab, init=False):
        """
        Construct a GeditTabDecorator

        @param window_decorator: the parent GeditWindowDecorator
        @param tab: the GeditTab to create this for
        @param init: has to be True if this is created on plugin init to decorate
                        already opened files
        """
        self._window_decorator = window_decorator
        self._tab = tab
        self._text_buffer = tab.get_document()
        self._text_view = tab.get_view()

        self._editor = None
        self._file = None

        # initially check the editor instance
        #
        # this needs to be done, because when we init for already opened files
        # (when plugin is activated in config) we get no "loaded" signal
        if init:
            self._adjust_editor()

        # listen to GeditDocument signals
        self._signals_handlers = [
                self._text_buffer.connect("loaded", self._on_load),
                self._text_buffer.connect("saved", self._on_save)
        ]

        LOG.debug("created %s" % self)

    @property
    def tab(self):
        return self._tab

    def _on_load(self, document):
        """
        A file has been loaded
        """
        LOG.debug("file loaded")

        self._adjust_editor()

    def _on_save(self, document):
        """
        The file has been saved
        """
        LOG.debug("saved")

        if not self._adjust_editor():
            # if the editor has not changed
            if self._editor:
                self._editor.on_save()

    def _adjust_editor(self):
        """
        Check if the URI has changed and manage Editor object according to
        file extension

        @return: True if the editor has changed
        """
        location = self._text_buffer.get_location()
        if location is None:
            # this happends when the plugin is activated in a running Gedit
            # and this decorator is created for the empty file

            LOG.debug("no file loaded")

            if self._window_decorator.window.get_active_view() == self._text_view:
                GObject.idle_add(self._window_decorator.adjust, self)

        else:
            file = File(location.get_uri())

            if file == self._file:        # FIXME: != doesn't work for File...
                return False
            else:
                LOG.debug("adjust_editor: URI has changed")

                self._file = file

                # URI has changed - manage the editor instance
                if self._editor:
                    # editor is present - destroy editor
                    self._editor.destroy()
                    self._editor = None

                # FIXME: comparing file extensions should be case-INsensitive...
                extension = file.extension

                # find Editor class for extension
                editor_class = None
                for clazz in EDITORS:
                    if extension in clazz.extensions:
                        editor_class = clazz
                        break

                if not editor_class is None:
                    # create instance
                    self._editor = editor_class.__new__(editor_class)
                    editor_class.__init__(self._editor, self, file)

                    # The following doesn't work because the right expression is evaluated
                    # and then assigned to the left. This means that Editor.__init__ is
                    # running and reading _editor while _editor is None. That leads to
                    #
                    # Traceback (most recent call last):
                    #   File "/home/michael/.gnome2/Gedit/plugins/GeditLaTeXPlugin/src/base/decorators.py", line 662, in _on_load
                    #     self._adjust_editor()
                    #   File "/home/michael/.gnome2/Gedit/plugins/GeditLaTeXPlugin/src/base/decorators.py", line 716, in _adjust_editor
                    #     self._editor = editor_class(self, file)
                    #   File "/home/michael/.gnome2/Gedit/plugins/GeditLaTeXPlugin/src/base/__init__.py", line 353, in __init__
                    #     self.init(file, self._window_context)
                    #   File "/home/michael/.gnome2/Gedit/plugins/GeditLaTeXPlugin/src/latex/editor.py", line 106, in init
                    #     self.__parse()
                    #   File "/home/michael/.gnome2/Gedit/plugins/GeditLaTeXPlugin/src/latex/editor.py", line 279, in __parse
                    #     self._outline_view.set_outline(self._outline)
                    #   File "/home/michael/.gnome2/Gedit/plugins/GeditLaTeXPlugin/src/latex/views.py", line 228, in set_outline
                    #     OutlineConverter().convert(self._store, outline, self._offset_map, self._context.active_editor.edited_file)

                    #self._editor = editor_class(self, file)
                else:
                    LOG.info("No editor class found for extension %s" % extension)

                # tell WindowDecorator to adjust actions
                # but only if this tab is the active tab
                if self._window_decorator.window.get_active_view() == self._text_view:
                    GObject.idle_add(self._window_decorator.adjust,self)

                # notify that URI has changed
                return True

    @property
    def file(self):
        return self._file

    @property
    def editor(self):
        return self._editor

    @property
    def extension(self):
        """
        @return: the extension of the currently opened file
        """
        if self._file is None:
            return None
        else:
            return self._file.extension

    def destroy(self):
        # disconnect from signals
        for handler in self._signals_handlers:
            self._text_buffer.disconnect(handler)

        # unreference the window decorator
        del self._window_decorator

        # destroy Editor instance
        if not self._editor is None:
            self._editor.destroy()

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
base.completion
"""

from logging import getLogger
from gi.repository import GObject, Gtk, Gdk, GdkPixbuf


from .preferences import Preferences


class ICompletionHandler(object):
    """
    This should be implemented for each language or 'proposal source'
    """
    @property
    def trigger_keys(self):
        """
        @return: a list of gdk key codes that trigger completion
        """
        raise NotImplementedError

    @property
    def prefix_delimiters(self):
        """
        @return: a list of characters that delimit the prefix on the left
        """
        raise NotImplementedError

    def complete(self, prefix):
        """
        @return: a list of objects extending Proposal
        """
        raise NotImplementedError


class Proposal(object):
    """
    A proposal for completion
    """
    @property
    def source(self):
        """
        @return: a subclass of Source to be inserted on activation
        """
        raise NotImplementedError

    @property
    def label(self):
        """
        @return: a string (may be pango markup) to be shown in proposals popup
        """
        raise NotImplementedError

    @property
    def details(self):
        """
        @return: a widget to be shown in details popup
        """
        raise NotImplementedError

    @property
    def icon(self):
        """
        @return: an instance of GdkPixbuf.Pixbuf
        """
        raise NotImplementedError

    @property
    def overlap(self):
        """
        @return: the number of overlapping characters from the beginning of the
            proposal and the prefix it was generated for
        """
        raise NotImplementedError

    def __key__(self):
        return self.label.lower()

    def __lt__(self, other):
        return self.__key__() < other.__key__()

    def __eq__(self, other):
        return self.__key__() == other.__key__()

    def __hash__(self):
        return hash(self.__key__())

class ProposalPopup(Gtk.Window):
    """
    Popup showing a list of proposals. This is implemented as a singleton
    as it doesn't make sense to have multiple popups around.
    """

    _log = getLogger("ProposalPopup")

    _POPUP_WIDTH = 300
    _POPUP_HEIGHT = 200
    _SPACE = 0

    def __new__(cls):
        if not '_instance' in cls.__dict__:
            cls._instance = Gtk.Window.__new__(cls)
        return cls._instance

    def __init__(self):
        if not '_ready' in dir(self):
            Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)
            #self, Gtk.WindowType.POPUP)

            self._store = Gtk.ListStore(str, object, GdkPixbuf.Pixbuf)        # markup, Proposal instance

            self._view = Gtk.TreeView(model=self._store)

            # pack the icon and text cells in one column to avoid the column separator
            column = Gtk.TreeViewColumn()
            pixbuf_renderer = Gtk.CellRendererPixbuf()
            column.pack_start(pixbuf_renderer, False)
            column.add_attribute(pixbuf_renderer, "pixbuf", 2)

            text_renderer = Gtk.CellRendererText()
            column.pack_start(text_renderer, True)
            column.add_attribute(text_renderer, "markup", 0)

            self._view.append_column(column)

#            self._view.insert_column_with_attributes(-1, "", Gtk.CellRendererPixbuf(), pixbuf=2)
#            self._view.insert_column_with_attributes(-1, "", Gtk.CellRendererText(), markup=0)

            self._view.set_enable_search(False)
            self._view.set_headers_visible(False)

            scr = Gtk.ScrolledWindow()
            scr.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            scr.add(self._view)
            scr.set_size_request(self._POPUP_WIDTH, self._POPUP_HEIGHT)

            frame = Gtk.Frame()
            frame.set_shadow_type(Gtk.ShadowType.OUT)
            frame.add(scr)

            self.add(frame)

            self._details_popup = DetailsPopup()

            self._ready = True

    @property
    def selected_proposal(self):
        """
        Returns the currently selected proposal
        """
        store, it = self._view.get_selection().get_selected()
        return store.get_value(it, 1)

    def activate(self, proposals, text_view):
        """
        Load proposals, move to the cursor position and show
        """
        self._set_proposals(proposals)
        self._move_to_cursor(text_view)

        self.show_all()

        self._update_details_popup()

    def deactivate(self):
        """
        Hide this popup and the DetailsPopup
        """
        self._details_popup.deactivate()
        self.hide()

    def _set_proposals(self, proposals):
        """
        Loads proposals into the popup
        """
        # sort
        proposals.sort()

        # load
        self._store.clear()
        for proposal in proposals:
            self._store.append([proposal.label, proposal, proposal.icon])

        self._view.set_cursor(Gtk.TreePath.new_from_string("0"), None, False)

    def navigate(self, key):
        """
        Moves the selection in the view according to key
        """
        if key == "Up":
            d = -1
        elif key == "Down":
            d = 1
        elif key == "Page_Up":
            d = -5
        elif key == "Page_Down":
            d = 5
        else:
            return

        path,column = self._view.get_cursor()
        index = int(path.to_string())
        max = self._store.iter_n_children(None)

        index += d

        if index < 0:
            index = 0
        elif index >= max:
            index = max - 1

        self._view.set_cursor(
                Gtk.TreePath.new_from_string("%d" % index),
                column,
                False)

        self._update_details_popup()

    def _get_cursor_pos(self, text_view):
        """
        Retrieve the current absolute position of the cursor in a TextView
        """
        buffer = text_view.get_buffer()
        location = text_view.get_iter_location(buffer.get_iter_at_mark(buffer.get_insert()))

        winX, winY = text_view.buffer_to_window_coords(Gtk.TextWindowType.WIDGET, location.x, location.y)

        win = text_view.get_window(Gtk.TextWindowType.WIDGET)
        ignore, xx, yy = win.get_origin()

        x = winX + xx
        y = winY + yy + location.height + self._SPACE

        return (x, y)

    def _update_details_popup(self):
        """
        Move and show the DetailsPopup if the currently selected proposal
        contains details.
        """
        try:
            path,column = self._view.get_cursor()
            index = int(path.to_string())

            proposal = self._store[index][1]

#            self._log.debug("proposal.details: " + str(proposal.details))

            if proposal.details is None:
                self._details_popup.hide()
                return

            # move
            x, y = self.get_position()
            width = self.get_size()[0]
            path, column = self._view.get_cursor()
            rect = self._view.get_cell_area(path, column)

            self._details_popup.move(x + width + 2, y + rect.y)

            # activate
            self._details_popup.activate(proposal.details)

        except Exception as e:
            self._log.error(e)

    def _move_to_cursor(self, text_view):
        """
        Move the popup to the current location of the cursor
        """
        sw = Gdk.Screen.width()
        sh = Gdk.Screen.height()

        x, y = self._get_cursor_pos(text_view)

        w, h = self.get_size()

        if x + w > sw:
            x = sw - w - 4

        if y + h > sh:
            # get the height of a character
            layout = text_view.create_pango_layout("a")
            ytext = layout.get_pixel_size()[1]
            y = y - ytext - h

        self.move(x, y)


class DetailsPopup(Gtk.Window):
    """
    A popup showing additional information at the right of the currently
    selected proposal in the ProposalPopup.

    This is used to display details of a BibTeX entry or the result of
    a template.
    """

    def __init__(self):
        Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)

        self._color = Preferences().get("light-foreground-color")

        self._label = Gtk.Label()
        self._label.set_use_markup(True)
        self._label.set_alignment(0, .5)

        self._frame = Gtk.Frame()
        self._frame.set_shadow_type(Gtk.ShadowType.OUT)
        #self._frame.set_border_width(3)
        self._frame.add(self._label)

        self.add(self._frame)

    def activate(self, details):
        """
        Create widget(s) for the given details and show the popup
        """
        # remove the old child widget if present
        child = self._frame.get_child()
        if child:
            self._frame.remove(child)
            child.destroy()

        # create a child widget depending on the type of details
        if type(details) is list:
            # table data
            table = Gtk.Table()
            table.set_border_width(5)
            table.set_col_spacings(5)
            rc = 0
            for row in details:
                cc = 0
                for column in row:
                    if cc == 0:
                        # first column
                        label = Gtk.Label("<span color='%s'>%s</span>" % (self._color, column))
                    else:
                        label = Gtk.Label(label=column)
                    label.set_use_markup(True)
                    if cc == 0:
                        # 1st column is right aligned
                        label.set_alignment(1.0, 0.5)
                    else:
                        # others are left aligned
                        label.set_alignment(0.0, 0.5)
                    table.attach(label, cc, cc + 1, rc, rc + 1)
                    cc += 1
                rc += 1
            self._frame.add(table)

        else:
            # markup text
            label = Gtk.Label(label=details)
            label.set_use_markup(True)
            self._frame.add(label)

        self.show_all()

        # force a recompute of the window size
        self.resize(1, 1)

    def deactivate(self):
        """
        Hide the popup
        """
        self.hide()


class CompletionDistributor(object):
    """
    This forms the lower end of the completion mechanism and hosts one
    or more CompletionHandlers
    """

    # TODO: clearify and simplify states here!
    # TODO: auto-close (maybe...)

    _log = getLogger("CompletionDistributor")

    _MAX_PREFIX_LENGTH = 100

    # completion delay in ms
    _DELAY = 500

    _STATE_IDLE, _STATE_CTRL_PRESSED, _STATE_ACTIVE = 0, 1, 2

    # keys that abort completion
    _ABORT_KEYS = ["Escape", "Left", "Right", "Home", "End", "space", "Tab"]

    # keys that are used to navigate in the popup
    _NAVIGATION_KEYS = ["Up", "Down", "Page_Up", "Page_Down"]

    # some characters have key constants that differ from their value
    _SPECIAL_KEYS = {"@": "at"}

    def __init__(self, editor, handlers):
        """
        @param editor: the instance of Editor this CompletionDistributor should observe
        @param handlers: a list of ICompletionHandler instances
        """

        self._log.debug("init")

        self._handlers = handlers        # we already get objects

        self._editor = editor
        self._text_buffer = editor.tab_decorator.tab.get_document()
        self._text_view = editor.tab_decorator.tab.get_view()

        self._state = self._STATE_IDLE
        self._timer = None

        # collect trigger keys from all handlers
        self._trigger_keys = []
        for handler in self._handlers:
            for key in handler.trigger_keys:
                if key in list(self._SPECIAL_KEYS.keys()):
                    self._trigger_keys.append(self._SPECIAL_KEYS[key])
                else:
                    self._trigger_keys.append(key)

        # TODO: is it right to instatiate this here?
        self._popup = ProposalPopup()

        # connect to signals
        self._signal_handlers = [
                self._text_view.connect("key-press-event", self._on_key_pressed),
                self._text_view.connect_after("key-release-event", self._on_key_released),
                self._text_view.connect("button-press-event", self._on_button_pressed),
                self._text_view.connect("focus-out-event", self._on_focus_out)
        ]

    def _on_key_pressed(self, view, event):
        """
        """
        key = Gdk.keyval_name(event.keyval)

        if self._state == self._STATE_IDLE:
            if key == "Control_L" or key == "Control_R":
                self._state = self._STATE_CTRL_PRESSED

            elif key in self._trigger_keys:
                self._start_timer()
                return False

        elif self._state == self._STATE_ACTIVE:
            if key == "Return":
                # select proposal

                self._abort()

                proposal = self._popup.selected_proposal
                self._select_proposal(proposal)

                # TODO: self._autoClose(proposalSelected=True)

                # returning True stops the signal
                return True

            elif key in self._ABORT_KEYS:
                self._abort()

            elif key in self._NAVIGATION_KEYS:
                self._popup.navigate(key)

                # returning True stops the signal
                return True

        elif self._state == self._STATE_CTRL_PRESSED:
            if key == "space":
                self._state = self._STATE_IDLE
                self._complete()
                return True
            elif key != "Control_L" and key != "Control_R":
                self._state = self._STATE_IDLE

        self._stop_timer()

    def _on_key_released(self, view, event):
        """
        """
        key = Gdk.keyval_name(event.keyval)

#        # trigger auto close on "}"
#        if key == "braceright":
#            # TODO: self._autoClose(braceTyped=True)
#            pass

        if self._state == self._STATE_ACTIVE:
            if key in self._NAVIGATION_KEYS or key in ["Control_L", "Control_R", "space"]:
                # returning True stops the signal
                return True
            else:
                # user is typing on with active popup...

                # TODO: we should check here if the cursor has moved
                # or better if the buffer has changed

                self._complete()

    def _on_button_pressed(self, view, event):
        self._abort()

    def _on_focus_out(self, view, event):
        self._abort()

    def _start_timer(self):
        """
        Start a timer for completion
        """
        # stop eventually running timer
        self._stop_timer()
        # start timer
        self._timer = GObject.timeout_add(self._DELAY, self._timer_callback)

    def _stop_timer(self):
        """
        Stop the timer if it has been started
        """
        if self._timer is not None:
            GObject.source_remove(self._timer)
            self._timer = None

    def _timer_callback(self):
        """
        Timeout
        """
        self._complete()

        return False    # do not call again

    def _complete(self):
        all_proposals = []

        for handler in self._handlers:
            delimiters = handler.prefix_delimiters

            prefix = self._find_prefix(delimiters)
            if prefix:
#                if handler.strip_delimiter:
#                    prefix = prefix[1:]

                proposals = handler.complete(prefix)
                assert type(proposals) is list

                all_proposals.extend(proposals)

        if len(all_proposals):
            self._popup.activate(all_proposals, self._text_view)
            self._state = self._STATE_ACTIVE
        else:
            self._abort()

    def _find_prefix(self, delimiters):
        """
        Find the start of the surrounding command and return the text
        from there to the cursor position.

        This is the prefix forming the basis for LaTeX completion.
        """
        it_right = self._text_buffer.get_iter_at_mark(self._text_buffer.get_insert())
        it_left = it_right.copy()

        # go back by one character (insert iter points to the char at the right of
        # the cursor!)
        if not it_left.backward_char():
            self._log.debug("_find_prefix: start of buffer reached")
            return None

        # move left until 'the left-most of a sequence of delimiters'
        #
        # e.g. if 'x' is a delimiter and 'abcxxx' is at left of the cursor, we
        # recognize 'xxx' as the prefix instead of only 'x'
        delim_found = False
        delim_char = None

        i = 0
        while i < self._MAX_PREFIX_LENGTH:
            c = it_left.get_char()

            if delim_found:
                if c != delim_char:
                    # a delimiter has been found and the preceding character
                    # is different from it
                    break
            else:
                if c in delimiters:
                    # a delimiter has been found
                    delim_found = True
                    delim_char = c

            if not it_left.backward_char():
                self._log.debug("_find_prefix: start of buffer reached")
                return None

            i += 1

        # to recognize the left-most delimiter, we have moved one char
        # too much
        it_left.forward_char()

        if i == self._MAX_PREFIX_LENGTH:
            self._log.debug("_find_prefix: prefix too long")
            return None

        prefix = self._text_buffer.get_text(it_left, it_right, False)

        return prefix

    def _select_proposal(self, proposal):
        """
        Insert the source contained in the activated proposal
        """
        self._editor.delete_at_cursor(- proposal.overlap)
        self._editor.insert(proposal.source)

    def _abort(self):
        """
        Abort completion
        """
        if self._state == self._STATE_ACTIVE:
            self._popup.deactivate()
            self._state = self._STATE_IDLE

    def destroy(self):
        # unreference the editor (very important! cyclic reference)
        del self._editor

        # disconnect text view events
        for handler in self._signal_handlers:
            self._text_view.disconnect(handler)

    def __del__(self):
        self._log.debug("Properly destroyed %s" % self)


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.dialogs
"""

import logging
import tempfile
import os.path
import string

from gi.repository import Gtk, GdkPixbuf

from ..preferences import Preferences
from ..util import GladeInterface
from ..resources import Resources
from ..file import File, Folder

from .preview import PreviewRenderer, ImageToolGenerator
from .environment import Environment
from .listing import LanguagesParser

from . import LaTeXSource

class ComboBoxProxy:
    """
    This proxies a ComboBox widget:

    p = ComboBoxProxy(self.find_widget("myCombo"), "SomeSetting")
    p.add_option("thing_1", "First Thing")
    p.add_option("thing_2", "Second Thing")
    p.restore("thing_1")
    ...
    """
    def __init__(self, widget, key):
        self._widget = widget
        self._key = key
        self._preferences = Preferences()

        self._store = Gtk.ListStore(str, str)            # value, label
        self._widget.set_model(self._store)
        cell = Gtk.CellRendererText()
        self._widget.pack_start(cell, True)
        self._widget.add_attribute(cell, "markup", 1)

        self._options = []

    def restore(self, default):
        if default != None:
            restored_value = default
        else:
            restored_value = self._preferences.get(self._key)
        restored_index = 0
        i = 0
        for value, label in self._options:
            if value == restored_value:
                restored_index = i
                break
            i += 1
        self._widget.set_active(restored_index)

        self._widget.connect("changed", self._on_changed)

    def _on_changed(self, combobox):
        self._preferences.set(self._key, self.value)

    def add_option(self, value, label, show_value=True):
        """
        Add an option to the widget

        @param value: a unique value
        @param label: a label text that may contain markup
        """
        self._options.append((value, label))

        if show_value:
            if not value is None and len(value) > 0:
                label_markup = "%s <span color='%s'>%s</span>" % (value, self._preferences.get("light-foreground-color"), _(label))
            else:
                label_markup = "<span color='%s'>%s</span>" % (self._preferences.get("light-foreground-color"), _(label))
        else:
            label_markup = _(label)

        self._store.append([value, label_markup])

    @property
    def value(self):
        index = self._widget.get_active()
        return self._options[index][0]

class ChooseMasterDialog(GladeInterface):
    """
    Dialog for choosing a master file to a LaTeX fragment file
    """

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("choose_master_dialog.ui")

    def run(self, folder):
        """
        Runs the dialog and returns the selected filename

        @param folder: a folder to initially place the file chooser
        """
        dialog = self.find_widget("dialogSelectMaster")
        file_chooser_button = self.find_widget("filechooserbutton")
        file_chooser_button.set_current_folder(folder)

        if dialog.run() == 1:
            filename = file_chooser_button.get_filename()
        else:
            filename = None
        dialog.hide()

        return filename

class NewDocumentDialog(GladeInterface):
    """
    Dialog for creating the body of a new LaTeX document
    """

    _log = logging.getLogger("NewDocumentWizard")

    _PAPER_SIZES = (
        ("a4paper", "A4"),
        ("a5paper", "A5"),
        ("b5paper", "B5"),
        ("executivepaper", "US-Executive"),
        ("legalbl_paper", "US-Legal"),
        ("letterpaper", "US-Letter") )

    _DEFAULT_FONT_FAMILIES =  (
        ("\\rmdefault", "<span font_family=\"serif\">Roman</span>"),
        ("\\sfdefault", "<span font_family=\"sans\">Sans Serif</span>"),
        ("\\ttdefault", "<span font_family=\"monospace\">Typerwriter</span>") )

    _LOCALE_MAPPINGS = {
        "af"    : "afrikaans",
        "af_ZA" : "afrikaans",
        "sq"    : "albanian",
        "sq_AL" : "albanian",
        "eu"    : "basque",
        "eu_ES" : "basque",
        "id"    : "bahasai",
        "id_ID" : "bahasai",
        "ms"    : "bahasam",
        "ms_MY" : "bahasam",
        "br"    : "breton",
        "bg"    : "bulgarian",
        "bg_BG" : "bulgarian",
        "ca"    : "catalan",
        "ca_ES" : "catalan",
        "hr"    : "croatian",
        "hr_HR" : "croatian",
        "cz"    : "czech",
        "da"    : "danish",
        "da_DK" : "danish",
        "nl"    : "dutch",
        "nl_BE" : "dutch",
        "nl_NL" : "dutch",
        "eo"    : "esperanto",
        "et"    : "estonian",
        "et_EE" : "estonian",
        "en"    : "USenglish",
        "en_AU" : "australian",
        "en_CA" : "canadian",
        "en_GB" : "UKenglish",
        "en_US" : "USenglish",
        "en_ZA" : "UKenglish",
        "fi"    : "finnish",
        "fi_FI"    : "finnish",
        "fr"    : "frenchb",
        "fr_FR" : "frenchb",
        "fr_CA" : "canadien",
        "gl"    : "galician",
        "gl_ES" : "galician",
        "el"    : "greek",
        "el_GR" : "greek",
        "he"    : "hebrew",
        "he_IL" : "hebrew",
        "hu"    : "magyar",
        "hu_HU" : "magyar",
        "is"    : "icelandic",
        "is_IS" : "icelandic",
        "it"    : "italian",
        "it_CH" : "italian",
        "it_IT" : "italian",
        "ga"    : "irish",
        "la"    : "latin",
        "dsb"    : "lsorbian",
        "de_DE" : "ngermanb",
        "de_AT" : "naustrian",
        "de_CH" : "ngermanb",
        "nb"    : "norsk",
        "no"    : "norsk",
        "no_NO" : "norsk",
        "no_NY" : "norsk",
        "nn"    : "nynorsk",
        "nn_NO" : "nynorsk",
        "pl"    : "polish",
        "pl_PL" : "polish",
        "pt"    : "portuguese",
        "pt_BR" : "brazilian",
        "pt_PT" : "portuguese",
        "ro"    : "romanian",
        "ro_RO" : "romanian",
        "ru"    : "russianb",
        "ru_RU" : "russianb",
        "gd"    : "scottish",
        "se"    : "samin",
        "se_NO" : "samin",
        "sr"    : "serbian",
        "sr_YU" : "serbian",
        "sl"    : "slovene",
        "sl_SL" : "slovene",
        "sl_SI"    : "slovene",
        "sk"    : "slovak",
        "sk_SK" : "slovak",
        "es"    : "spanish",
        "es_AR" : "spanish",
        "es_CL" : "spanish",
        "es_CO" : "spanish",
        "es_DO" : "spanish",
        "es_EC" : "spanish",
        "es_ES" : "spanish",
        "es_GT" : "spanish",
        "es_HN" : "spanish",
        "es_LA" : "spanish",
        "es_MX" : "spanish",
        "es_NI" : "spanish",
        "es_PA" : "spanish",
        "es_PE" : "spanish",
        "es_PR" : "spanish",
        "es_SV" : "spanish",
        "es_UY" : "spanish",
        "es_VE" : "spanish",
        "es_UY" : "spanish",
        "sv"    : "swedish",
        "sv_SE" : "swedish",
        "sv_SV" : "swedish",
        "tr"    : "turkish",
        "tr_TR"    : "turkish",
        "uk"    : "ukraineb",
        "uk_UA" : "ukraineb",
        "hsb"    : "usorbian",
        "cy"    : "welsh",
        "cy_GB" : "welsh"
    }

    dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("new_document_template_dialog.ui")

    def get_dialog(self):
        """
        Build and return the dialog
        """
        if self.dialog == None:
            preferences = Preferences()
            environment = Environment()

            self.dialog = self.find_widget("dialogNewDocument")

            #
            # file
            #
            self._entry_name = self.find_widget("entryName")
            self._button_dir = self.find_widget("buttonDirectory")
            self._button_dir.set_action(Gtk.FileChooserAction.SELECT_FOLDER)

            #
            # templates
            #
            self._proxy_template = ComboBoxProxy(self.find_widget("comboTemplate"), "RecentTemplate")

            folder = Folder(preferences.TEMPLATE_DIR)

            templates = folder.files
            default_template_path = Resources().get_data_file("templates")
            folder = Folder(default_template_path)
            templates.extend(folder.files)
            templates.sort()
            for template in templates:
                self._proxy_template.add_option(template.path, template.shortbasename, show_value=False)
            self._proxy_template.restore(os.path.join(default_template_path,"Default.template"))

            #
            # metadata
            #
            self._entry_title = self.find_widget("entryTitle")
            self._entry_author = self.find_widget("entryAuthor")
            self._entry_author.set_text(environment.username)
            self._radio_date_custom = self.find_widget("radioCustom")
            self._entry_date = self.find_widget("entryDate")

            #
            # document classes
            #
            self._proxy_document_class = ComboBoxProxy(self.find_widget("comboClass"), "RecentDocumentClass")
            document_classes = environment.document_classes
            document_classes.sort(key=lambda x: x.name.lower())
            for c in document_classes:
                self._proxy_document_class.add_option(c.name, _(c.label))
            self._proxy_document_class.restore("article")

            #
            # paper
            #
            self._proxy_paper_size = ComboBoxProxy(self.find_widget("comboPaperSize"), "RecentPaperSize")
            self._proxy_paper_size.add_option("", _("Default"))
            for size, label in self._PAPER_SIZES:
                self._proxy_paper_size.add_option(size, label)
            self._proxy_paper_size.restore("")


            self._check_landscape = self.find_widget("checkLandscape")
            self._check_landscape.set_active(False)

            #
            # font size
            #
            self._radio_font_user = self.find_widget("radioFontUser")
            self._spin_font_size = self.find_widget("spinFontSize")
            self._labelFontSize = self.find_widget("labelFontSize")

            #
            # font family
            #
            self._proxy_font_family = ComboBoxProxy(self.find_widget("comboFontFamily"), "RecentDefaultFontFamily")
            for command, label in self._DEFAULT_FONT_FAMILIES:
                self._proxy_font_family.add_option(command, _(label), False)
            self._proxy_font_family.restore("\\rmdefault")


            #
            # input encodings
            #
            self._proxy_encoding = ComboBoxProxy(self.find_widget("comboEncoding"), "RecentInputEncoding")
            input_encodings = environment.input_encodings
            input_encodings.sort(key=lambda x: x.name.lower())
            for e in input_encodings:
                self._proxy_encoding.add_option(e.name, e.label)
            self._proxy_encoding.restore("utf8")

            #
            # babel packages
            #
            self._proxy_babel = ComboBoxProxy(self.find_widget("comboBabel"), "RecentBabelPackage")
            language_definitions = environment.language_definitions
            language_definitions.sort(key=lambda x: x.name.lower())
            for l in language_definitions:
                self._proxy_babel.add_option(l.name, _(l.label))

            try:
                self._proxy_babel.restore(self._LOCALE_MAPPINGS[environment.language_code])
            except Exception as e:
                self._log.error("Failed to guess babel package: %s" % e)
                self._proxy_babel.restore("english")

            #
            # connect signals
            #
            self.connect_signals({ "on_radioCustom_toggled" : self._on_custom_date_toggled,
                                   "on_radioFontUser_toggled" : self._on_user_font_toggled })

        return self.dialog

    def _on_custom_date_toggled(self, toggle_button):
        self._entry_date.set_sensitive(toggle_button.get_active())

    def _on_user_font_toggled(self, toggle_button):
        self._spin_font_size.set_sensitive(toggle_button.get_active())
        self._labelFontSize.set_sensitive(toggle_button.get_active())

    @property
    def source(self):
        """
        Compose and return the source resulting from the dialog
        """
        # document class options
        documentOptions = []

        if self._radio_font_user.get_active():
            documentOptions.append("%spt" % self._spin_font_size.get_value_as_int())

#        paperSize = self._store_paper_size[self._combo_paper_size.get_active()][0]
#        if len(paperSize) > 0:
#            documentOptions.append(paperSize)
        paperSize = self._proxy_paper_size.value
        if paperSize != "":
            documentOptions.append(paperSize)

        if self._check_landscape.get_active():
            documentOptions.append("landscape")

        if len(documentOptions) > 0:
            documentOptions = "[" + ",".join(documentOptions) + "]"
        else:
            documentOptions = ""


#        documentClass = self._store_class[self._combo_class.get_active()][0]
        documentClass = self._proxy_document_class.value

        title = self._entry_title.get_text()
        author = self._entry_author.get_text()
#        babelPackage = self._store_babel[self._combo_babel.get_active()][0]
#        inputEncoding = self._store_encoding[self._combo_encoding.get_active()][0]
        babelPackage = self._proxy_babel.value
        inputEncoding = self._proxy_encoding.value

        if self._radio_date_custom.get_active():
            date = self._entry_date.get_text()
        else:
            date = "\\today"

        if self.find_widget("checkPDFMeta").get_active():
            ifpdf = "\n\\usepackage{ifpdf}\n\\usepackage{hyperref}"

            # pdfinfo is discouraged because it
            #  - doesn't support special characters like umlauts
            #  - is not supported by XeTeX
            # see http://sourceforge.net/tracker/index.php?func=detail&aid=2809478&group_id=204144&atid=988431

#            pdfinfo = """
#\\ifpdf
#    \\pdfinfo {
#        /Author (%s)
#        /Title (%s)
#        /Subject (SUBJECT)
#        /Keywords (KEYWORDS)
#        /CreationDate (D:%s)
#    }
#\\fi""" % (author, title, time.strftime('%Y%m%d%H%M%S'))

            pdfinfo = """
\\ifpdf
\\hypersetup{
    pdfauthor={%s},
    pdftitle={%s},
}
\\fi""" % (author, title)
        else:
            ifpdf = ""
            pdfinfo = ""

        if self._proxy_font_family.value == "\\rmdefault":
            default_font_family = ""    # \rmdefault is the default value of \familydefault
        else:
            default_font_family = "\n\\renewcommand{\\familydefault}{%s}" % self._proxy_font_family.value

        template_string = open(self._proxy_template.value).read()
        template = string.Template(template_string)
        s = template.safe_substitute({
                    "DocumentOptions" : documentOptions,
                    "DocumentClass" : documentClass,
                    "InputEncoding" : inputEncoding,
                    "BabelPackage" : babelPackage,
                    "AdditionalPackages" : default_font_family + ifpdf,
                    "Title" : title,
                    "Author" : author,
                    "Date" : date,
                    "AdditionalPreamble" : pdfinfo})

#        s = """\\documentclass%s{%s}
#\\usepackage[%s]{inputenc}
#\\usepackage[%s]{babel}%s%s
#\\title{%s}
#\\author{%s}
#\\date{%s}%s
#\\begin{document}
#
#\\end{document}""" % (documentOptions, documentClass, inputEncoding, babelPackage, default_font_family, ifpdf, title, author, date, pdfinfo)

        return s

    @property
    def file(self):
        """
        Return the File object
        """
        return File("%s/%s.tex" % (self._button_dir.get_filename(), self._entry_name.get_text()))

    def run(self):
        """
        Runs the dialog
        """
        dialog = self.get_dialog()
        r = dialog.run()
        dialog.hide()
        return r


class UseBibliographyDialog(GladeInterface, PreviewRenderer):
    """
    Dialog for inserting a reference to a bibliography
    """

    _log = logging.getLogger("UseBibliographyWizard")


    # sample bibtex content used for rendering the preview
    _BIBTEX = """@book{ dijkstra76,
    title={{A Discipline of Programming}},
    author={Edsger W. Dijkstra},
    year=1976 }
@article{ dijkstra68,
    title={{Go to statement considered harmful}},
    author={Edsger W. Dijkstra},
    year=1968 }"""

    dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("use_bibliography_dialog.ui")

    def run_dialog(self, edited_file):
        """
        Run the dialog

        @param edited_file: the File edited the active Editor
        @return: the source resulting from the dialog
        """
        source = None
        dialog = self._get_dialog()

        self._file_chooser_button.set_current_folder(edited_file.dirname)

        if dialog.run() == 1:        # TODO: use gtk constant
            base_dir = edited_file.dirname

            file = File(self._file_chooser_button.get_filename())
            source = "\\bibliography{%s}\n\\bibliographystyle{%s}" % (file.relativize_shortname(base_dir),
                                                                    self._storeStyle[self._comboStyle.get_active()][0])

        dialog.hide()

        return source

    def _get_dialog(self):
        if not self.dialog:

            self.dialog = self.find_widget("dialogUseBibliography")

            # bib file

            self._file_chooser_button = self.find_widget("filechooserbutton")

            # styles

            self._storeStyle = Gtk.ListStore(str)

            styles = Environment().bibtex_styles
            for style in styles:
                self._storeStyle.append([style.name])

            self._comboStyle = self.find_widget("comboboxStyle")
            self._comboStyle.set_model(self._storeStyle)

            try:
                recent = styles.index(Preferences().get("recent-bibtex-style"))
            except ValueError:
                recent = 0
            self._comboStyle.set_active(recent)

            self._imagePreview = self.find_widget("previewimage")
            self._imagePreview.show()

            self.connect_signals({ "on_buttonRefresh_clicked" : self._on_refresh_clicked })

        return self.dialog

    def _on_refresh_clicked(self, widget):
        """
        The button for refreshing the preview has been clicked
        """
        index = self._comboStyle.get_active()
        if index < 0:
            self._log.error("No style selected")
            return

        style = self._storeStyle[index][0]

        self._imagePreview.set_from_stock(Gtk.STOCK_EXECUTE, Gtk.IconSize.BUTTON)

        # create temporary bibtex file
        self._tempFile = tempfile.NamedTemporaryFile(mode="w", suffix=".bib")
        self._tempFile.write(self._BIBTEX)
        self._tempFile.flush()

        filename = self._tempFile.name
        self._filenameBase = os.path.splitext(os.path.basename(filename))[0]

        # build preview image
        self.render("Book \\cite{dijkstra76} Article \\cite{dijkstra68} \\bibliography{%s}\\bibliographystyle{%s}" % (self._filenameBase,
                                                                                                                    style))
    def _on_render_succeeded(self, pixbuf):
        # PreviewRenderer._on_render_succeeded
        self._imagePreview.set_from_pixbuf(pixbuf)
        # remove the temp bib file
        self._tempFile.close()

    def _on_render_failed(self):
        # PreviewRenderer._on_render_failed

        # set a default icon as preview
        self._imagePreview.set_from_stock(Gtk.STOCK_STOP, Gtk.IconSize.BUTTON)
        # remove the temp bib file
        self._tempFile.close()


class InsertGraphicsDialog(GladeInterface):

    _PREVIEW_WIDTH, _PREVIEW_HEIGHT = 128, 128
    _dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("insert_graphics_dialog.ui")

    def run(self, edited_file):
        """
        @param edited_file: the File currently edited
        """
        dialog = self.__get_dialog()

        source = None

        if dialog.run() == 1:
            #filename = self._fileChooser.get_filename()

            file = File(self._fileChooser.get_filename())
            relative_filename = file.relativize(edited_file.dirname)

            # for eps and pdf the extension should be omitted
            # (see http://www.tex.ac.uk/cgi-bin/texfaq2html?label=graph-pspdf)
            ext = os.path.splitext(relative_filename)[1]
            if ext == ".pdf" or ext == ".eps":
                relative_filename = os.path.splitext(relative_filename)[0]

            width = "%.2f" % round(self.find_widget("spinbuttonWidth").get_value() / 100.0, 2)
            caption = self.find_widget("entryCaption").get_text()
            label = self.find_widget("entryLabel").get_text()
            floating = self.find_widget("checkbuttonFloat").get_active()
            spread = self.find_widget("checkbuttonSpread").get_active()
            relativeTo = self._comboRelative.get_active()
            rotate = self.find_widget("spinRotate").get_value()
            flip = self.find_widget("checkFlip").get_active()

            source = ""
            packages = ["graphicx"]

            if relativeTo == 0:        # relative to image size
                options = "scale=%s" % width
            else:        # relative to text width
                options = "width=%s\\textwidth" % width

            if rotate != 0:
                options += ", angle=%s" % rotate

            includegraphics = "\\includegraphics[%s]{%s}" % (options, relative_filename)

            if flip:
                includegraphics = "\\reflectbox{%s}" % includegraphics

            if floating:
                if spread:
                    ast = "*"
                else:
                    ast = ""
                source += "\\begin{figure%s}[ht]\n\t\\centering\n\t%s" % (ast, includegraphics)
                if len(caption):
                    source += "\n\t\\caption{%s}" % caption
                if len(label) and label != "fig:":
                    source += "\n\t\\label{%s}" % label
                source += "\n\\end{figure%s}" % ast
            else:
                source += "\\begin{center}\n\t%s" % includegraphics
                if len(caption):
                    source += "\n\t\\captionof{figure}{%s}" % caption
                    packages.append("caption")
                if len(label) and label != "fig:":
                    source += "\n\t\\label{%s}" % label
                source += "\n\\end{center}"

#            viewAdapter.insertText(source)
#            viewAdapter.ensurePackages(packages)

            source = LaTeXSource(source, packages)

        dialog.hide()

        return source

    def __get_dialog(self):
        if not self._dialog:

            # setup the dialog

            self._dialog = self.find_widget("dialogInsertGraphics")

            previewImage = Gtk.Image()
            self._fileChooser = self.find_widget("FileChooser")
            self._fileChooser.set_preview_widget(previewImage)
            self._fileChooser.connect("update-preview", self.__update_preview, previewImage)

            #self._fileChooser.get_property("dialog").connect("response", self._fileChooserResponse)    # FIXME: not readable
            #self._fileChooser.connect("file-activated", self._fileActivated)                            # FIXME: doesn't work

            self._okayButton = self.find_widget("buttonOkay")

            self._comboRelative = self.find_widget("comboRelative")
            self._comboRelative.set_active(0)

        return self._dialog

    def __update_preview(self, fileChooser, previewImage):
        """
        Update the FileChooser's preview image
        """
        filename = fileChooser.get_preview_filename()

        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(filename, self._PREVIEW_WIDTH, self._PREVIEW_HEIGHT)
            previewImage.set_from_pixbuf(pixbuf)
            fileChooser.set_preview_widget_active(True)
        except:
            fileChooser.set_preview_widget_active(False)


class InsertTableDialog(GladeInterface):
    """
    This is used to include tables and matrices
    """

    _dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("insert_table_dialog.ui")

    def run(self):
        dialog = self.__get_dialog()

        source = None

        if dialog.run() == 1:
            floating = self.find_widget("checkbuttonFloat").get_active()
            rows = int(self.find_widget("spinbuttonRows").get_value())
            cols = int(self.find_widget("spinbuttonColumns").get_value())
            caption = self.find_widget("entryCaption").get_text()
            label = self.find_widget("entryLabel").get_text()

            s = ""

            if self.find_widget("radiobuttonTable").get_active():
                # table

                layout = "l" * cols

                if floating:
                    s += "\\begin{table}[ht]\n\t\\centering\n\t\\begin{tabular}{%s}%s\n\t\\end{tabular}" % (layout,
                                                                                        self.__build_table_body(rows, cols, "\t\t"))

                    if len(caption):
                        s += "\n\t\\caption{%s}" % caption

                    if len(label) and label != "tab:":
                        s += "\n\t\\label{%s}" % label

                    s += "\n\\end{table}"

                else:
                    s += "\\begin{center}\n\t\\begin{tabular}{%s}%s\n\t\\end{tabular}" % (layout,
                                                                                        self.__build_table_body(rows, cols, "\t\t"))

                    if len(caption):
                        s += "\n\t\\captionof{table}{%s}" % caption

                    if len(label):
                        s += "\n\t\\label{%s}" % label

                    s += "\n\\end{center}"
                packages = []
            else:
                environ = self._storeDelims[self._comboDelims.get_active()][2]
                s = "\\begin{%s}%s\n\\end{%s}" % (environ, self.__build_table_body(rows, cols, "\t\t"), environ)
                packages = ["amsmath"]

            source = LaTeXSource(s, packages)

        dialog.hide()

        return source

    def __get_dialog(self):
        if not self._dialog:
            self._dialog = self.find_widget("dialogInsertTable")

            # delimiters
            self._storeDelims = Gtk.ListStore(GdkPixbuf.Pixbuf, str, str)      # icon, label, environment

            self._storeDelims.append([None, "None", "matrix"])

            delimiters = [("parantheses", "Parantheses", "pmatrix"),
                        ("brackets", "Brackets", "bmatrix"),
                        ("braces", "Braces", "Bmatrix"),
                        ("vbars", "Vertical Bars", "vmatrix"),
                        ("dvbars", "Double Vertical Bars", "Vmatrix")]

            for d in delimiters:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("%s.png" % d[0]))
                self._storeDelims.append([pixbuf, d[1], d[2]])

            self._comboDelims = self.find_widget("comboDelims")
            self._comboDelims.set_model(self._storeDelims)

            cellPixbuf = Gtk.CellRendererPixbuf()
            self._comboDelims.pack_start(cellPixbuf, False)
            self._comboDelims.add_attribute(cellPixbuf, 'pixbuf', 0)

            cellText = Gtk.CellRendererText()
            self._comboDelims.pack_start(cellText, False)
            self._comboDelims.add_attribute(cellText, 'text', 1)

            self._comboDelims.set_active(0)

            self.connect_signals({ "on_radioMatrix_toggled" : self.__matrix_toggled })

        return self._dialog

    def __matrix_toggled(self, toggleButton):
        self._comboDelims.set_sensitive(toggleButton.get_active())

    def __build_table_body(self, rows, cols, indent):
        """
        This builds the body of a table template according to the number of rows and
        columns.
        """

        # FIXME: create unique placeholders for each cell for multi-placeholder feature

        s = ""
        for i in range(rows):
            colList = []
            for j in range(cols):
                colList.append("${%s%s}" % (i + 1, j + 1))
            s += "\n" + indent + " & ".join(colList) + " \\\\\\\\"
        return s


class InsertListingDialog(GladeInterface):
    """
    """

    _dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("insert_listing_dialog.ui")

    def run(self, edited_file):
        """
        @param edited_file: the File currently edited
        """
        dialog = self.__get_dialog()

        source = None

        if dialog.run() == 1:

            lstset = ""
            options = ""
            extra = r',basicstyle=\small\ttfamily,breaklines=true,showtabs=false,showspaces=false,breakatwhitespace=true'

            language = self._storeLanguages[self._comboLanguage.get_active()][0]
            try:
                dialect = self._storeDialects[self._comboDialect.get_active()][0]
            except IndexError:
                dialect = ""

            if self._dialectsEnabled and len(dialect):
                # we need the lstset command
                lstset = "\\lstset{language=[%s]%s%s}\n" % (dialect, language, extra)
            else:
                options = "[language=%s%s]" % (language, extra)


            if self._checkFile.get_active():
                file = File(self._fileChooserButton.get_filename())
                relative_filename = file.relativize(edited_file.dirname)

                source = "%s\\lstinputlisting%s{%s}" % (lstset, options, relative_filename)

            else:
                source = "%s\\begin{lstlisting}%s\n\t$0\n\\end{lstlisting}" % (lstset, options)

            source = LaTeXSource(source, ["listings"])

        dialog.hide()

        return source

    def __get_dialog(self):
        if not self._dialog:
            self._dialog = self.find_widget("dialogListing")

            self._fileChooserButton = self.find_widget("fileChooserButton")

            #
            # languages
            #
            self._languages = []
            parser = LanguagesParser()
            parser.parse(self._languages, Resources().get_data_file("listings.xml"))

            recentLanguage = Preferences().get("recent-listing-language")

            self._storeLanguages = Gtk.ListStore(str)
            recentLanguageIndex = 0
            i = 0
            for l in self._languages:
                self._storeLanguages.append([l.name])
                if l.name == recentLanguage:
                    recentLanguageIndex = i
                i += 1

            self._comboLanguage = self.find_widget("comboLanguage")
            self._comboLanguage.set_model(self._storeLanguages)
            cell = Gtk.CellRendererText()
            self._comboLanguage.pack_start(cell, True)
            self._comboLanguage.add_attribute(cell, "text", 0)

            #
            # dialects
            #
            self._labelDialect = self.find_widget("labelDialect")

            self._storeDialects = Gtk.ListStore(str)

            self._comboDialect = self.find_widget("comboDialect")
            self._comboDialect.set_model(self._storeDialects)
            cell = Gtk.CellRendererText()
            self._comboDialect.pack_start(cell, True)
            self._comboDialect.add_attribute(cell, "text", 0)

            self._checkFile = self.find_widget("checkFile")

            self.connect_signals({ "on_checkFile_toggled" : self._fileToggled,
                                  "on_comboLanguage_changed" : self._languagesChanged })


            self._comboLanguage.set_active(recentLanguageIndex)

            self._dialectsEnabled = False

        return self._dialog

    def _languagesChanged(self, comboBox):
        language = self._languages[comboBox.get_active()]

        self._storeDialects.clear()

        if len(language.dialects):
            i = 0
            for dialect in language.dialects:
                self._storeDialects.append([dialect.name])
                if dialect.default:
                    self._comboDialect.set_active(i)
                i += 1

            self._labelDialect.set_sensitive(True)
            self._comboDialect.set_sensitive(True)

            self._dialectsEnabled = True
        else:
            self._labelDialect.set_sensitive(False)
            self._comboDialect.set_sensitive(False)

            self._dialectsEnabled = False

    def _fileToggled(self, toggleButton):
        self._fileChooserButton.set_sensitive(toggleButton.get_active())

    def _specificToggled(self, toggleButton):
        self._comboLanguage.set_sensitive(toggleButton.get_active())
        self._labelDialect.set_sensitive(toggleButton.get_active() and self._dialectsEnabled)
        self._comboDialect.set_sensitive(toggleButton.get_active() and self._dialectsEnabled)


class BuildImageDialog(GladeInterface):
    """
    Render the document to an image
    """

    _dialog = None
    _generator = ImageToolGenerator()

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("build_image_dialog.ui")

    def run(self):
        dialog = self._getDialog()

        if dialog.run() == 1:

            if self.find_widget("radioPNG").get_active():
                self._generator.format = ImageToolGenerator.FORMAT_PNG
                self._generator.pngMode = self._storeMode[self._comboMode.get_active()][1]
            elif self.find_widget("radioJPEG").get_active():
                self._generator.format = ImageToolGenerator.FORMAT_JPEG
            elif self.find_widget("radioGIF").get_active():
                self._generator.format = ImageToolGenerator.FORMAT_GIF

            self._generator.open = True
            self._generator.resolution = self._spinResolution.get_value_as_int()
            self._generator.antialiasFactor = self._storeAntialias[self._comboAntialias.get_active()][1]
            self._generator.render_box = self.find_widget("radioBox").get_active()

            generate = True
        else:
            generate = False

        dialog.hide()

        if generate:
            return self._generator.generate()
        else:
            return None

    def _getDialog(self):
        if not self._dialog:
            self._dialog = self.find_widget("dialogRenderImage")
            self.find_widget("grid1").set_column_spacing(20)

            # PNG mode

            self._storeMode = Gtk.ListStore(str, int)    # label, mode constant
            self._storeMode.append([_("Monochrome"), ImageToolGenerator.PNG_MODE_MONOCHROME])
            self._storeMode.append([_("Grayscale"), ImageToolGenerator.PNG_MODE_GRAYSCALE])
            self._storeMode.append([_("RGB"), ImageToolGenerator.PNG_MODE_RGB])
            self._storeMode.append([_("RGBA"), ImageToolGenerator.PNG_MODE_RGBA])

            self._comboMode = self.find_widget("comboMode")
            self._comboMode.set_model(self._storeMode)
            cell = Gtk.CellRendererText()
            self._comboMode.pack_start(cell, True)
            self._comboMode.add_attribute(cell, "text", 0)
            self._comboMode.set_active(3)

            # anti-alias

            self._storeAntialias = Gtk.ListStore(str, int)    # label, factor
            self._storeAntialias.append([_("Off"), 0])
            self._storeAntialias.append([_("1x"), 1])
            self._storeAntialias.append([_("2x"), 2])
            self._storeAntialias.append([_("4x"), 4])
            self._storeAntialias.append([_("8x"), 8])

            self._comboAntialias = self.find_widget("comboAntialias")
            self._comboAntialias.set_model(self._storeAntialias)
            cell = Gtk.CellRendererText()
            self._comboAntialias.pack_start(cell, True)
            self._comboAntialias.add_attribute(cell, "text", 0)
            self._comboAntialias.set_active(3)

            # resolution

            self._spinResolution = self.find_widget("spinResolution")
            self._spinResolution.set_value(Environment().screen_dpi)

        return self._dialog


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.lexer
"""

class StringListener(object):
    """
    Recognizes a string in a stream of characters
    """
    def __init__(self, string, any_position=True):
        """
        @param string: the character sequence to be recognized
        @param any_position: if True the sequence may occur at any position in
                the stream, if False it must occur at the start
        """
        self._string = string
        self._last = len(string)
        self._pos = 0
        self._any_position = any_position

        self._active = True

    def put(self, char):
        """
        Returns True if the string is recognized
        """
        if not self._active:
            return False

        if char == self._string[self._pos]:
            self._pos += 1

            if self._pos == self._last:
                return True
        else:
            if self._any_position:
                self._pos = 0
            else:
                self._active = False

        return False


from ..util import StringReader


class Token(object):
    """
    A Token returned by the Lexer
    """

    COMMAND, TEXT, COMMENT, VERBATIM, BEGIN_CURLY, END_CURLY, BEGIN_SQUARE, END_SQUARE = list(range(8))

    def __init__(self, type, offset=None, value=None):
        self.type = type
        self.offset = offset
        self.value = value

    @property
    def xml(self):
        if self.type == self.COMMAND:
            return "<t:command>%s</t:command>" % self.value
        elif self.type == self.TEXT:
            return "<t:text>%s</t:text>" % self.value
        elif self.type == self.VERBATIM:
            return "<t:verbatim>%s</t:verbatim>" % self.value
        elif self.type == self.COMMENT:
            return "<t:comment>%s</t:comment>" % self.value
        else:
            return "<t:terminal />"


class Lexer(object):
    """
    LaTeX lexer
    """

    # TODO: redesign and optimize this from a DFA

    # states of the lexer
    _DEFAULT, _BACKSLASH, _COMMAND, _TEXT, _COMMENT, _PRE_VERB, _VERB, _VERBATIM = list(range(8))

    _SPECIAL = set(["&", "$", "{", "}", "[", "]", "%", "#", "_", "\\"])

    _TERMINALS = set(["{", "}", "[", "]"])
    _TERMINALS_MAP = {"{" : Token.BEGIN_CURLY, "}" : Token.END_CURLY,
                      "[" : Token.BEGIN_SQUARE, "]" : Token.END_SQUARE}

    _VERBATIM_ENVIRONS = set(["verbatim", "verbatim*", "lstlisting", "lstlisting*"])


    # additional states for recognizing "\begin{verbatim}"
    _VERBATIM_DEFAULT, _VERBATIM_BEGIN, _VERBATIM_BEGIN_CURLY, _VERBATIM_BEGIN_CURLY_ENVIRON = list(range(4))


    def __init__(self, string, skipWs=True, skipComment=False):
        self._reader = StringReader(string)

        self._skipWs = skipWs
        self._skipComment = skipComment

        self._state = self._DEFAULT
        self._verbatimState = self._VERBATIM_DEFAULT

        self._eof = False
        self._tokenStack = []    # used to return a sequence of tokens after a verbatim ended

    def __iter__(self):
        return self

    def __next__(self):
        if self._eof:
            raise StopIteration

        # first empty the token stack
        if len(self._tokenStack):
            return self._tokenStack.pop()

        while True:
            try:
                char = self._reader.read()

                if self._state == self._DEFAULT:
                    if char == "\\":
                        self._state = self._BACKSLASH
                        self._verbatimState = self._VERBATIM_DEFAULT
                        self._startOffset = self._reader.offset - 1

                    elif char == "%":
                        self._state = self._COMMENT
                        self._verbatimState = self._VERBATIM_DEFAULT
                        self._startOffset = self._reader.offset - 1
                        if not self._skipComment:
                            self._text = []

                    elif char in self._TERMINALS:
                        if self._verbatimState == self._VERBATIM_BEGIN and char == "{":
                            self._verbatimState = self._VERBATIM_BEGIN_CURLY

                        elif self._verbatimState == self._VERBATIM_BEGIN_CURLY_ENVIRON and char == "}":
                            # we have "\begin{verbatim}"
                            self._verbatimState = self._VERBATIM_DEFAULT
                            self._state = self._VERBATIM
                            self._text = []
                            self._startOffset = self._reader.offset
                            self._verbatimSequenceListener = StringListener("\\end{%s}" % self._verbatimEnviron)

                        else:
                            self._verbatimState = self._VERBATIM_DEFAULT

                        return Token(self._TERMINALS_MAP[char], self._reader.offset - 1)

                    else:
                        self._state = self._TEXT
                        self._startOffset = self._reader.offset - 1
                        self._text = [char]

                elif self._state == self._BACKSLASH:
                    if char in self._SPECIAL or char.isspace():
                        # this is a one-character-command, also whitespace is allowed
                        self._state = self._DEFAULT
                        return Token(Token.COMMAND, self._startOffset, char)

                    else:
                        self._state = self._COMMAND

                        self._verbListener = StringListener("verb", any_position=False)
                        self._verbListener.put(char)

                        self._text = [char]

                elif self._state == self._COMMENT:
                    if char == "\n":
                        self._state = self._DEFAULT
                        if not self._skipComment:
                            return Token(Token.COMMENT, self._startOffset, "".join(self._text))

                    else:
                        if not self._skipComment:
                            self._text.append(char)

                elif self._state == self._COMMAND:
                    if char in self._SPECIAL or char.isspace():

                        name = "".join(self._text)

                        # this is mostly false because \verb is mostly followed by something like |
                        if name == "verb":
                            self._state = self._VERB
                            self._verbDelimiter = char
                            self._startOffset = self._reader.offset - 1
                            self._text = [char]

                        elif name == "url":     # we handle "\url" just like "\verb"
                            self._state = self._VERB
                            self._verbDelimiter = "}"
                            self._startOffset = self._reader.offset - 1
                            self._text = []

                        else:
                            self._state = self._DEFAULT
                            self._reader.unread(char)

                            if name == "begin":
                                self._verbatimState = self._VERBATIM_BEGIN

                            return Token(Token.COMMAND, self._startOffset, name)

                    else:
                        if self._verbListener.put(char):
                            # we have "\verb"
                            self._state = self._PRE_VERB
                        else:
                            self._text.append(char)

                elif self._state == self._PRE_VERB:
                    self._state = self._VERB
                    self._verbDelimiter = char
                    self._startOffset = self._reader.offset - 1
                    self._text = []

                elif self._state == self._TEXT:
                    if char in self._SPECIAL:
                        self._state = self._DEFAULT
                        self._reader.unread(char)

                        text = "".join(self._text)

                        if self._skipWs and text.isspace():
                            continue
                        else:

                            if self._verbatimState == self._VERBATIM_BEGIN_CURLY:
                                # we have "\begin{" until now, handle verbatim environment

                                if text in self._VERBATIM_ENVIRONS:
                                    self._verbatimEnviron = text
                                    self._verbatimState = self._VERBATIM_BEGIN_CURLY_ENVIRON

                                else:
                                    self._verbatimState = self._VERBATIM_DEFAULT

                            return Token(Token.TEXT, self._startOffset, text)

                    else:
                        self._text.append(char)

                elif self._state == self._VERB:
                    if char == self._verbDelimiter:        # FIXME: \overbrace
                        self._state = self._DEFAULT

                        return Token(Token.VERBATIM, self._startOffset, "".join(self._text) + char)

                    else:
                        self._text.append(char)

                elif self._state == self._VERBATIM:
                    if self._verbatimSequenceListener.put(char):
                        self._state = self._DEFAULT

                        # TODO: calculate offsets
                        self._tokenStack = [ Token(Token.END_CURLY, 0),
                                             Token(Token.TEXT, 0, self._verbatimEnviron),
                                             Token(Token.BEGIN_CURLY, 0),
                                             Token(Token.COMMAND, 0, "end") ]

                        text = "".join(self._text)
                        text = text[5:]        # cut off "\end{"
                        return Token(Token.VERBATIM, self._startOffset, text)
                    else:
                        self._text.append(char)

                elif self._state == self._VERB:
                    # this char is the verb delimiter
                    # TODO: implement verbatim detection
                    pass

            except StopIteration:
                self._eof = True

                # evaluate final state
                if self._state == self._BACKSLASH:
                    return Token(Token.COMMAND, self._startOffset, "")

                elif self._state == self._COMMAND:
                    return Token(Token.COMMAND, self._startOffset, "".join(self._text))

                elif self._state == self._TEXT:
                    text = "".join(self._text)
                    if not (self._skipWs and text.isspace()):
                        return Token(Token.TEXT, self._startOffset, text)

                elif self._state == self._VERB:

                    # TODO: the document is malformed in this case, so the lexer should be
                    # able to produce issues, too
                    #
                    # TODO: return a VERBATIM token

                    return Token(Token.TEXT, self._startOffset, "".join(self._text))

                raise StopIteration
# ex:ts=4:et:
